@@ -5,14 +5,18 @@ import asyncio
 import time
 import webbrowser
 from threading import Lock
-from typing import Callable, Optional, Coroutine
+from typing import Callable, Coroutine, List, Optional, Any
 
 from azure.core.exceptions import ClientAuthenticationError
-from azure.identity import ManagedIdentityCredential, AzureCliCredential
+from azure.core.tracing.decorator import distributed_trace
+from azure.core.tracing.decorator_async import distributed_trace_async
+from azure.core.tracing import SpanKind
+from azure.identity import AzureCliCredential, ManagedIdentityCredential
 from msal import ConfidentialClientApplication, PublicClientApplication
 
-from ._cloud_settings import CloudSettings
-from .exceptions import KustoClientError, KustoAioSyntaxError, KustoAsyncUsageError
+from ._cloud_settings import CloudInfo, CloudSettings
+from ._telemetry import KustoTracing
+from .exceptions import KustoAioSyntaxError, KustoAsyncUsageError, KustoClientError
 
 try:
     from asgiref.sync import sync_to_async
@@ -23,13 +27,32 @@ except ImportError:
 
 
 try:
-    from azure.identity.aio import ManagedIdentityCredential as AsyncManagedIdentityCredential, AzureCliCredential as AsyncAzureCliCredential
-except ImportError:
+    from azure.identity.aio import (
+        ManagedIdentityCredential as AsyncManagedIdentityCredential,
+        AzureCliCredential as AsyncAzureCliCredential,
+        DefaultAzureCredential as AsyncDefaultAzureCredential,
+    )
 
+    from azure.core.credentials_async import AsyncTokenCredential
+except ImportError:
+    # These are here in case the user doesn't have the aio optional dependency installed, but still tries to use async.
+    # They will give them a useful error message, and will appease linters.
     class AsyncManagedIdentityCredential:
         def __init__(self):
             raise KustoAioSyntaxError()
 
+    class AsyncAzureCliCredential:
+        def __init__(self):
+            raise KustoAioSyntaxError()
+
+    class AsyncDefaultAzureCredential:
+        def __init__(self):
+            raise KustoAioSyntaxError()
+
+    class AsyncTokenCredential:
+        def __init__(self):
+            raise KustoAioSyntaxError()
+
 
 # constant key names and values used throughout the code
 class TokenConstants:
@@ -54,22 +77,22 @@ class TokenProviderBase(abc.ABC):
     The class is build for Lazy initialization, so that the first call, take on instantiation of 'heavy' long-lived class members
     """
 
-    _initialized = False
-    _cloud_initialized = False
-    _kusto_uri = None
-    _cloud_info = None
-    _scopes = None
+    _initialized: bool = False
+    _resources_initialized: bool = False
 
-    def __init__(self, kusto_uri: str, is_async: bool = False):
+    def __init__(self, is_async: bool = False):
+        self._proxy_dict: Optional[str, str] = None
         self.is_async = is_async
-        self._kusto_uri = kusto_uri
 
         if is_async:
             self._async_lock = asyncio.Lock()
         else:
             self._lock = Lock()
 
-    def _init_once(self, init_only_cloud=False):
+    def close(self):
+        pass
+
+    def _init_once(self, init_only_resources=False):
         if self._initialized:
             return
 
@@ -77,17 +100,17 @@ class TokenProviderBase(abc.ABC):
             if self._initialized:
                 return
 
-            if not self._cloud_initialized:
-                self._init_cloud()
-                self._cloud_initialized = True
+            if not self._resources_initialized:
+                self._init_resources()
+                self._resources_initialized = True
 
-            if init_only_cloud:
+            if init_only_resources:
                 return
 
             self._init_impl()
             self._initialized = True
 
-    async def _init_once_async(self, init_only_cloud=False):
+    async def _init_once_async(self, init_only_resources=False):
         if self._initialized:
             return
 
@@ -95,66 +118,73 @@ class TokenProviderBase(abc.ABC):
             if self._initialized:
                 return
 
-            if not self._cloud_initialized:
-                await (sync_to_async(self._init_cloud)())
-                self._cloud_initialized = True
+            if not self._resources_initialized:
+                await sync_to_async(self._init_resources)()
+                self._resources_initialized = True
 
-            if init_only_cloud:
+            if init_only_resources:
                 return
 
             self._init_impl()
             self._initialized = True
 
-    def _init_cloud(self):
-        if self._kusto_uri is not None:
-            self._cloud_info = CloudSettings.get_cloud_info_for_cluster(self._kusto_uri)
-            resource_uri = self._cloud_info.kusto_service_resource_id
-            if self._cloud_info.login_mfa_required:
-                resource_uri = resource_uri.replace(".kusto.", ".kustomfa.")
-
-            self._scopes = [resource_uri + "/.default"]
+    def _init_resources(self):
+        pass
 
     def get_token(self):
         """Get a token silently from cache or authenticate if cached token is not found"""
-        if self.is_async:
-            raise KustoAsyncUsageError("get_token", self.is_async)
-        self._init_once()
 
-        token = self._get_token_from_cache_impl()
-        if token is None:
-            with self._lock:
-                token = self._get_token_impl()
+        @distributed_trace(name_of_span=f"{self.name()}.get_token", tracing_attributes=self.context(), kind=SpanKind.CLIENT)
+        def _get_token():
+            if self.is_async:
+                raise KustoAsyncUsageError("get_token", self.is_async)
+            self._init_once()
 
-        return self._valid_token_or_throw(token)
+            token = self._get_token_from_cache_impl()
+            if token is None:
+                with self._lock:
+                    token = KustoTracing.call_func_tracing(
+                        self._get_token_impl, name_of_span=f"{self.name()}.get_token_impl", tracing_attributes=self.context()
+                    )
+
+            return self._valid_token_or_throw(token)
+
+        return _get_token()
 
     def context(self) -> dict:
         if self.is_async:
             raise KustoAsyncUsageError("context", self.is_async)
-        self._init_once(init_only_cloud=True)
+        self._init_once(init_only_resources=True)
         return self._context_impl()
 
     async def context_async(self) -> dict:
         if not self.is_async:
             raise KustoAsyncUsageError("context_async", self.is_async)
 
-        await self._init_once_async(init_only_cloud=True)
+        await self._init_once_async(init_only_resources=True)
         return self._context_impl()
 
     async def get_token_async(self):
         """Get a token asynchronously silently from cache or authenticate if cached token is not found"""
 
-        if not self.is_async:
-            raise KustoAsyncUsageError("get_token_async", self.is_async)
+        @distributed_trace_async(name_of_span=f"{self.name()}.get_token_async", tracing_attributes=self.context_async(), kind=SpanKind.CLIENT)
+        async def _get_token_async():
+            if not self.is_async:
+                raise KustoAsyncUsageError("get_token_async", self.is_async)
 
-        await self._init_once_async()
+            await self._init_once_async()
 
-        token = self._get_token_from_cache_impl()
+            token = self._get_token_from_cache_impl()
 
-        if token is None:
-            async with self._async_lock:
-                token = await self._get_token_impl_async()
+            if token is None:
+                async with self._async_lock:
+                    token = await KustoTracing.call_func_tracing_async(
+                        self._get_token_impl_async, name_of_span=f"{self.name()}.get_token_impl_async", tracing_attributes=self.context_async()
+                    )
 
-        return self._valid_token_or_throw(token)
+            return self._valid_token_or_throw(token)
+
+        return await _get_token_async()
 
     @staticmethod
     @abc.abstractmethod
@@ -205,12 +235,34 @@ class TokenProviderBase(abc.ABC):
 
         return token
 
+    def set_proxy(self, proxy_url: str):
+        self._proxy_dict = {"http": proxy_url, "https": proxy_url}
+
+
+class CloudInfoTokenProvider(TokenProviderBase, abc.ABC):
+    _cloud_info: Optional[CloudInfo]
+    _scopes = List[str]
+    _kusto_uri: str
+
+    def __init__(self, kusto_uri: str, is_async: bool = False):
+        super().__init__(is_async)
+        self._kusto_uri = kusto_uri
+
+    def _init_resources(self):
+        if self._kusto_uri is not None:
+            self._cloud_info = CloudSettings.get_cloud_info_for_cluster(self._kusto_uri, self._proxy_dict)
+            resource_uri = self._cloud_info.kusto_service_resource_id
+            if self._cloud_info.login_mfa_required:
+                resource_uri = resource_uri.replace(".kusto.", ".kustomfa.")
+
+            self._scopes = [resource_uri + "/.default"]
+
 
 class BasicTokenProvider(TokenProviderBase):
     """Basic Token Provider keeps and returns a token received on construction"""
 
     def __init__(self, token: str, is_async: bool = False):
-        super().__init__(None, is_async)
+        super().__init__(is_async)
         self._token = token
 
     @staticmethod
@@ -223,7 +275,7 @@ class BasicTokenProvider(TokenProviderBase):
     def _init_impl(self):
         pass
 
-    def _get_token_impl(self) -> dict:
+    def _get_token_impl(self) -> Optional[dict]:
         return None
 
     def _get_token_from_cache_impl(self) -> dict:
@@ -236,7 +288,7 @@ class CallbackTokenProvider(TokenProviderBase):
     def __init__(
         self, token_callback: Optional[Callable[[], str]], async_token_callback: Optional[Callable[[], Coroutine[None, None, str]]], is_async: bool = False
     ):
-        super().__init__(None, is_async)
+        super().__init__(is_async)
         self._token_callback = token_callback
         self._async_token_callback = async_token_callback
 
@@ -257,7 +309,7 @@ class CallbackTokenProvider(TokenProviderBase):
 
         return {TokenConstants.MSAL_TOKEN_TYPE: TokenConstants.BEARER_TYPE, TokenConstants.MSAL_ACCESS_TOKEN: caller_token}
 
-    def _get_token_impl(self) -> dict:
+    def _get_token_impl(self) -> Optional[dict]:
         if self._token_callback is None:
             raise KustoClientError("token_callback is None, can't retrieve token")
         return self._build_response(self._token_callback())
@@ -267,11 +319,11 @@ class CallbackTokenProvider(TokenProviderBase):
             return await super()._get_token_impl_async()
         return self._build_response(await self._async_token_callback())
 
-    def _get_token_from_cache_impl(self) -> dict:
+    def _get_token_from_cache_impl(self) -> Optional[dict]:
         return None
 
 
-class MsiTokenProvider(TokenProviderBase):
+class MsiTokenProvider(CloudInfoTokenProvider):
     """
     MSI Token Provider obtains a token from the MSI endpoint
     The args parameter is a dictionary conforming with the ManagedIdentityCredential initializer API arguments
@@ -295,12 +347,12 @@ class MsiTokenProvider(TokenProviderBase):
     def _init_impl(self):
         pass
 
-    def _get_token_impl(self) -> dict:
+    def _get_token_impl(self) -> Optional[dict]:
         try:
             if self._msi_auth_context is None:
                 self._msi_auth_context = ManagedIdentityCredential(**self._msi_args)
 
-            msi_token = self._msi_auth_context.get_token(self._kusto_uri)
+            msi_token = self._msi_auth_context.get_token(self._scopes[0])
             return {TokenConstants.MSAL_TOKEN_TYPE: TokenConstants.BEARER_TYPE, TokenConstants.MSAL_ACCESS_TOKEN: msi_token.token}
         except ClientAuthenticationError as e:
             raise KustoClientError("Failed to initialize MSI ManagedIdentityCredential with [{0}]\n{1}".format(self._msi_args, e))
@@ -312,18 +364,24 @@ class MsiTokenProvider(TokenProviderBase):
             if self._msi_auth_context_async is None:
                 self._msi_auth_context_async = AsyncManagedIdentityCredential(**self._msi_args)
 
-            msi_token = await self._msi_auth_context_async.get_token(self._kusto_uri)
+            msi_token = await self._msi_auth_context_async.get_token(self._scopes[0])
             return {TokenConstants.MSAL_TOKEN_TYPE: TokenConstants.BEARER_TYPE, TokenConstants.MSAL_ACCESS_TOKEN: msi_token.token}
         except ClientAuthenticationError as e:
             raise KustoClientError("Failed to initialize MSI async ManagedIdentityCredential with [{0}]\n{1}".format(self._msi_args, e))
         except Exception as e:
             raise KustoClientError("Failed to obtain MSI token for '{0}' with [{1}]\n{2}".format(self._kusto_uri, self._msi_args, e))
 
-    def _get_token_from_cache_impl(self) -> dict:
+    def _get_token_from_cache_impl(self) -> Optional[dict]:
         return None
 
+    def close(self):
+        if self._msi_auth_context is not None:
+            self._msi_auth_context.close()
+        if self._msi_auth_context_async is not None:
+            self._msi_auth_context_async.close()
 
-class AzCliTokenProvider(TokenProviderBase):
+
+class AzCliTokenProvider(CloudInfoTokenProvider):
     """AzCli Token Provider obtains a refresh token from the AzCli cache and uses it to authenticate with MSAL"""
 
     def __init__(self, kusto_uri: str, is_async: bool = False):
@@ -342,12 +400,12 @@ class AzCliTokenProvider(TokenProviderBase):
     def _init_impl(self):
         pass
 
-    def _get_token_impl(self) -> dict:
+    def _get_token_impl(self) -> Optional[dict]:
         try:
             if self._az_auth_context is None:
                 self._az_auth_context = AzureCliCredential()
 
-            self._az_token = self._az_auth_context.get_token(self._kusto_uri)
+            self._az_token = self._az_auth_context.get_token(self._scopes[0])
             return {TokenConstants.AZ_TOKEN_TYPE: TokenConstants.BEARER_TYPE, TokenConstants.AZ_ACCESS_TOKEN: self._az_token.token}
         except Exception as e:
             raise KustoClientError(
@@ -359,14 +417,14 @@ class AzCliTokenProvider(TokenProviderBase):
             if self._az_auth_context_async is None:
                 self._az_auth_context_async = AsyncAzureCliCredential()
 
-            self._az_token = await self._az_auth_context_async.get_token(self._kusto_uri)
+            self._az_token = await self._az_auth_context_async.get_token(self._scopes[0])
             return {TokenConstants.AZ_TOKEN_TYPE: TokenConstants.BEARER_TYPE, TokenConstants.AZ_ACCESS_TOKEN: self._az_token.token}
         except Exception as e:
             raise KustoClientError(
                 "Failed to obtain Az Cli token for '{0}'.\nPlease be sure AzCli version 2.3.0 and above is installed.\n{1}".format(self._kusto_uri, e)
             )
 
-    def _get_token_from_cache_impl(self) -> dict:
+    def _get_token_from_cache_impl(self) -> Optional[dict]:
         if self._az_token is not None:
             # A token is considered valid if it is due to expire in no less than 10 minutes
             cur_time = time.time()
@@ -375,8 +433,14 @@ class AzCliTokenProvider(TokenProviderBase):
 
         return None
 
+    def close(self):
+        if self._az_auth_context is not None:
+            self._az_auth_context.close()
+        if self._az_auth_context_async is not None:
+            self._az_auth_context_async.close()
 
-class UserPassTokenProvider(TokenProviderBase):
+
+class UserPassTokenProvider(CloudInfoTokenProvider):
     """Acquire a token from MSAL with username and password"""
 
     def __init__(self, kusto_uri: str, authority_id: str, username: str, password: str, is_async: bool = False):
@@ -394,9 +458,11 @@ class UserPassTokenProvider(TokenProviderBase):
         return {"authority": self._cloud_info.authority_uri(self._auth), "client_id": self._cloud_info.kusto_client_app_id, "username": self._user}
 
     def _init_impl(self):
-        self._msal_client = PublicClientApplication(client_id=self._cloud_info.kusto_client_app_id, authority=self._cloud_info.authority_uri(self._auth))
+        self._msal_client = PublicClientApplication(
+            client_id=self._cloud_info.kusto_client_app_id, authority=self._cloud_info.authority_uri(self._auth), proxies=self._proxy_dict
+        )
 
-    def _get_token_impl(self) -> dict:
+    def _get_token_impl(self) -> Optional[dict]:
         token = self._msal_client.acquire_token_by_username_password(username=self._user, password=self._pass, scopes=self._scopes)
         return self._valid_token_or_throw(token)
 
@@ -411,7 +477,7 @@ class UserPassTokenProvider(TokenProviderBase):
         return self._valid_token_or_none(token)
 
 
-class DeviceLoginTokenProvider(TokenProviderBase):
+class DeviceLoginTokenProvider(CloudInfoTokenProvider):
     """Acquire a token from MSAL with Device Login flow"""
 
     def __init__(self, kusto_uri: str, authority_id: str, device_code_callback=None, is_async: bool = False):
@@ -429,9 +495,11 @@ class DeviceLoginTokenProvider(TokenProviderBase):
         return {"authority": self._cloud_info.authority_uri(self._auth), "client_id": self._cloud_info.kusto_client_app_id}
 
     def _init_impl(self):
-        self._msal_client = PublicClientApplication(client_id=self._cloud_info.kusto_client_app_id, authority=self._cloud_info.authority_uri(self._auth))
+        self._msal_client = PublicClientApplication(
+            client_id=self._cloud_info.kusto_client_app_id, authority=self._cloud_info.authority_uri(self._auth), proxies=self._proxy_dict
+        )
 
-    def _get_token_impl(self) -> dict:
+    def _get_token_impl(self) -> Optional[dict]:
         flow = self._msal_client.initiate_device_flow(scopes=self._scopes)
         try:
             if self._device_code_callback:
@@ -458,10 +526,17 @@ class DeviceLoginTokenProvider(TokenProviderBase):
         return self._valid_token_or_none(token)
 
 
-class InteractiveLoginTokenProvider(TokenProviderBase):
+class InteractiveLoginTokenProvider(CloudInfoTokenProvider):
     """Acquire a token from MSAL with Device Login flow"""
 
-    def __init__(self, kusto_uri: str, authority_id: str, login_hint: Optional[str] = None, domain_hint: Optional[str] = None, is_async: bool = False):
+    def __init__(
+        self,
+        kusto_uri: str,
+        authority_id: str,
+        login_hint: Optional[str] = None,
+        domain_hint: Optional[str] = None,
+        is_async: bool = False,
+    ):
         super().__init__(kusto_uri, is_async)
         self._msal_client = None
         self._auth = authority_id
@@ -477,9 +552,11 @@ class InteractiveLoginTokenProvider(TokenProviderBase):
         return {"authority": self._cloud_info.authority_uri(self._auth), "client_id": self._cloud_info.kusto_client_app_id}
 
     def _init_impl(self):
-        self._msal_client = PublicClientApplication(client_id=self._cloud_info.kusto_client_app_id, authority=self._cloud_info.authority_uri(self._auth))
+        self._msal_client = PublicClientApplication(
+            client_id=self._cloud_info.kusto_client_app_id, authority=self._cloud_info.authority_uri(self._auth), proxies=self._proxy_dict
+        )
 
-    def _get_token_impl(self) -> dict:
+    def _get_token_impl(self) -> Optional[dict]:
         token = self._msal_client.acquire_token_interactive(
             scopes=self._scopes, prompt=TokenConstants.MSAL_INTERACTIVE_PROMPT, login_hint=self._login_hint, domain_hint=self._domain_hint
         )
@@ -495,7 +572,7 @@ class InteractiveLoginTokenProvider(TokenProviderBase):
         return self._valid_token_or_none(token)
 
 
-class ApplicationKeyTokenProvider(TokenProviderBase):
+class ApplicationKeyTokenProvider(CloudInfoTokenProvider):
     """Acquire a token from MSAL with application Id and Key"""
 
     def __init__(self, kusto_uri: str, authority_id: str, app_client_id: str, app_key: str, is_async: bool = False):
@@ -514,10 +591,10 @@ class ApplicationKeyTokenProvider(TokenProviderBase):
 
     def _init_impl(self):
         self._msal_client = ConfidentialClientApplication(
-            client_id=self._app_client_id, client_credential=self._app_key, authority=self._cloud_info.authority_uri(self._auth)
+            client_id=self._app_client_id, client_credential=self._app_key, authority=self._cloud_info.authority_uri(self._auth), proxies=self._proxy_dict
         )
 
-    def _get_token_impl(self) -> dict:
+    def _get_token_impl(self) -> Optional[dict]:
         token = self._msal_client.acquire_token_for_client(scopes=self._scopes)
         return self._valid_token_or_throw(token)
 
@@ -526,13 +603,22 @@ class ApplicationKeyTokenProvider(TokenProviderBase):
         return self._valid_token_or_none(token)
 
 
-class ApplicationCertificateTokenProvider(TokenProviderBase):
+class ApplicationCertificateTokenProvider(CloudInfoTokenProvider):
     """
     Acquire a token from MSAL using application certificate
     Passing the public certificate is optional and will result in Subject Name & Issuer Authentication
     """
 
-    def __init__(self, kusto_uri: str, client_id: str, authority_id: str, private_cert: str, thumbprint: str, public_cert: str = None, is_async: bool = False):
+    def __init__(
+        self,
+        kusto_uri: str,
+        client_id: str,
+        authority_id: str,
+        private_cert: str,
+        thumbprint: str,
+        public_cert: str = None,
+        is_async: bool = False,
+    ):
         super().__init__(kusto_uri, is_async)
         self._msal_client = None
         self._auth = authority_id
@@ -554,13 +640,54 @@ class ApplicationCertificateTokenProvider(TokenProviderBase):
 
     def _init_impl(self):
         self._msal_client = ConfidentialClientApplication(
-            client_id=self._client_id, client_credential=self._cert_credentials, authority=self._cloud_info.authority_uri(self._auth)
+            client_id=self._client_id, client_credential=self._cert_credentials, authority=self._cloud_info.authority_uri(self._auth), proxies=self._proxy_dict
         )
 
-    def _get_token_impl(self) -> dict:
+    def _get_token_impl(self) -> Optional[dict]:
         token = self._msal_client.acquire_token_for_client(scopes=self._scopes)
         return self._valid_token_or_throw(token)
 
     def _get_token_from_cache_impl(self) -> dict:
         token = self._msal_client.acquire_token_silent(scopes=self._scopes, account=None)
         return self._valid_token_or_none(token)
+
+
+class AzureIdentityTokenCredentialProvider(CloudInfoTokenProvider):
+    """Acquire a token using an Azure Identity credential"""
+
+    def __init__(
+        self,
+        kusto_uri: str,
+        is_async: bool = False,
+        credential: Optional[Any] = None,
+        credential_from_login_endpoint: Optional[Callable[[str], Any]] = None,
+    ):
+        super().__init__(kusto_uri, is_async)
+
+        self.credential = credential
+        self.credential_from_login_endpoint = credential_from_login_endpoint
+
+        if self.credential is None and self.credential_from_login_endpoint is None:
+            raise KustoClientError("Either a credential or a credential_from_login_endpoint must be provided")
+
+    @staticmethod
+    def name() -> str:
+        return "AzureIdentityTokenProvider"
+
+    def _context_impl(self) -> dict:
+        return {"credential": self.credential}
+
+    def _init_impl(self):
+        if self.credential is None:
+            self.credential = self.credential_from_login_endpoint(self._cloud_info.login_endpoint)
+
+    def _get_token_impl(self) -> Optional[dict]:
+        t = self.credential.get_token(self._scopes[0])
+        return {TokenConstants.MSAL_TOKEN_TYPE: TokenConstants.BEARER_TYPE, TokenConstants.MSAL_ACCESS_TOKEN: t.token}
+
+    async def _get_token_impl_async(self) -> Optional[dict]:
+        t = await self.credential.get_token(self._scopes[0])
+        return {TokenConstants.MSAL_TOKEN_TYPE: TokenConstants.BEARER_TYPE, TokenConstants.MSAL_ACCESS_TOKEN: t.token}
+
+    def _get_token_from_cache_impl(self) -> Optional[dict]:
+        return None
